/*
 * Created on 13-may-2005
 *
 * gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package com.iver.cit.gvsig.fmap.drivers.sde;

import java.awt.Component;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.sql.Types;
import java.util.Hashtable;
import java.util.TreeMap;
import java.util.Vector;

import com.esri.sde.sdk.client.SeColumnDefinition;
import com.esri.sde.sdk.client.SeConnection;
import com.esri.sde.sdk.client.SeDefs;
import com.esri.sde.sdk.client.SeException;
import com.esri.sde.sdk.client.SeExtent;
import com.esri.sde.sdk.client.SeFilter;
import com.esri.sde.sdk.client.SeLayer;
import com.esri.sde.sdk.client.SeObjectId;
import com.esri.sde.sdk.client.SeQuery;
import com.esri.sde.sdk.client.SeQueryInfo;
import com.esri.sde.sdk.client.SeRegisteredColumn;
import com.esri.sde.sdk.client.SeRow;
import com.esri.sde.sdk.client.SeShape;
import com.esri.sde.sdk.client.SeShapeFilter;
import com.esri.sde.sdk.client.SeSqlConstruct;
import com.esri.sde.sdk.client.SeTable;
import com.hardcode.gdbms.engine.data.DataSourceFactory;
import com.hardcode.gdbms.engine.data.driver.ObjectDriver;
import com.hardcode.gdbms.engine.data.edition.DataWare;
import com.hardcode.gdbms.engine.values.Value;
import com.hardcode.gdbms.engine.values.ValueFactory;
import com.iver.andami.PluginServices;
import com.iver.andami.messages.Messages;
import com.iver.andami.messages.NotificationManager;
import com.iver.cit.gvsig.fmap.DriverException;
import com.iver.cit.gvsig.fmap.DriverJdbcNotFoundExceptionType;
import com.iver.cit.gvsig.fmap.core.FShape;
import com.iver.cit.gvsig.fmap.core.ICanReproject;
import com.iver.cit.gvsig.fmap.core.IFeature;
import com.iver.cit.gvsig.fmap.core.IGeometry;
import com.iver.cit.gvsig.fmap.drivers.DBException;
import com.iver.cit.gvsig.fmap.drivers.DBLayerDefinition;
import com.iver.cit.gvsig.fmap.drivers.DriverAttributes;
import com.iver.cit.gvsig.fmap.drivers.FieldDescription;
import com.iver.cit.gvsig.fmap.drivers.IConnection;
import com.iver.cit.gvsig.fmap.drivers.IFeatureIterator;
import com.iver.cit.gvsig.fmap.drivers.db.utils.ConnectionWithParams;
import com.iver.cit.gvsig.fmap.drivers.db.utils.SingleVectorialDBConnectionManager;
import com.iver.cit.gvsig.fmap.edition.EditionException;
import com.iver.cit.gvsig.fmap.edition.IWriteable;
import com.iver.cit.gvsig.fmap.edition.IWriter;
import com.iver.cit.gvsig.fmap.layers.XMLException;
import com.iver.utiles.XMLEntity;
import com.iver.utiles.swing.JPasswordDlg;


/**
 * Driver of ArcSDE.
 * @author Vicente Caballero Navarro
 */
public class ArcSdeDriver implements ICanReproject, IWriteable, IVectorialSDEDriver, ObjectDriver{
    protected static Hashtable poolPassw = new Hashtable();
    private SeLayer layer;
    private SeQuery query=null;
    private SeQuery queryAux=null;
    private SeSqlConstruct sqlConstruct;
    private SeColumnDefinition[] colDefs;
    private SeRow row = null;
    private int numReg = -1;
    private Rectangle2D fullExtent = null;
    private Hashtable hashRelate = null;
    private int idSpatialColumn = -1;
    private String[] fields;
    private int[] fieldTypes;
    private long lastRowServed = -1;
//    private SeConnection alphanumericConnection = null;
    private IConnection spatialConnection=null;
    private String strEPSG = "-1";
    private String originalEPSG = "-1";
    private String sqlTotal;
    private long posActual = -1;
    private SeRow cachedRow = null;
    protected DBLayerDefinition lyrDef = null;
    protected Rectangle2D workingArea;
    protected String driverClass;
    protected String className;
    protected String catalogName;
	private ArcSdeWriter writer;
//	private String userName;
//	private String tableName;
	private String FIDfield;
	private String geometryField;
	private String whereClause;
	private String strSRID;
//	private String host;
//	private String port;
//	private String dbName;
//	private String connName;
//	private String dbUrl;
	private SeConnection alphanumericConnection;

    /**
     * Recorre el recordset creando una tabla Hash que usaremos para relacionar
     * el nmero de un registro con su identificador nico. Debe ser llamado
     * en el setData justo despus de crear el recorset principal
     */
    protected void doRelateID_FID() {
        hashRelate = new Hashtable();

        try {
            if (row == null) {
                System.out.println(" No rows fetched");

                return;
            }

            int index = 0;
            int fid=getLyrDef().getIdFieldID();
            Value value=getFieldValue(index,fid);

            while (value != null) {
                //SeShape shpVal = row.getShape(idSpatialColumn);
                //SeObjectId objID = shpVal.getFeatureId();
            	String theKey = value.toString();
                hashRelate.put(theKey, new Integer(index));
                index++;
                value = getFieldValue(index,fid);


            }

            numReg = index;
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
        	NotificationManager.addError(e);
		}
    }
    /**
	 * @return
	 */
    public Rectangle2D getFullExtent() {
		return fullExtent;
	}
    public IGeometry getShape(int index) {
        SeRow row;

        try {
            row = obtainRow(index);
            if (row==null)
            	return null;
            SeShape spVal = row.getShape(idSpatialColumn);
            IGeometry geom = ArcSdeFeatureIterator.getGeometry(spVal);
            return geom;
        } catch (SeException e) {
        	NotificationManager.addError(e);
        }
        return null;
    }

    public IFeatureIterator getFeatureIterator(String sql)
        throws DriverException {
        return null;
    }

    public IFeatureIterator getFeatureIterator(Rectangle2D r, String strEPSG)
        throws DriverException {
        /*
         *   Generate a rectangular shape that will be used as a filter
         */
        try {
            SeShape shape = new SeShape(layer.getCoordRef());
            SeExtent extent=layer.getExtent();
            shape.generateRectangle(extent);

            SeShape[] shapes = new SeShape[1];
            shapes[0] = shape;

            /*
             *   Retrieve all the shapes that are contained within the
             *   rectangles envelope.
             */
            SeShapeFilter[] filters = new SeShapeFilter[1];
            SeShapeFilter filter = null;

            filter = new SeShapeFilter(layer.getQualifiedName(),
                    layer.getSpatialColumn(), shape, SeFilter.METHOD_ENVP);
            filters[0] = filter;

            SeQuery spatialQuery = null;
            SeSqlConstruct sqlCons = new SeSqlConstruct(layer.getQualifiedName());

            spatialQuery = new SeQuery(alphanumericConnection, fields, sqlCons);

            /*
             *   Set spatial constraints
             */
            spatialQuery.setSpatialConstraints(SeQuery.SE_OPTIMIZE, false,
                filters);

            spatialQuery.prepareQuery();

            spatialQuery.execute();

            return new ArcSdeFeatureIterator(spatialQuery,getLyrDef().getFieldID());
        } catch (SeException e) {
        	NotificationManager.addError(e);
            return null;
        }
    }

    public void close() {
    }

    public void open() throws DriverException {
    }

    public int getRowIndexByFID(IFeature FID) {
        int resul;
        String strID = FID.getID();
        Integer rowIndex = (Integer) hashRelate.get(strID);

        if (rowIndex == null) {
            System.err.println("Error: No se ha encontrado " + strID +
                " dentro del Hash");
        }
        resul = rowIndex.intValue();
        return resul;
    }

    public String getGeometryField(String fieldName) {
        return layer.getSpatialColumn();
    }

    public int getShapeCount() throws IOException {
        return numReg;
    }

    public DriverAttributes getDriverAttributes() {
        return null;
    }

    public String getName() {
        return "gvSIG SDE driver";
    }

    /**
	 * @return
	 * @uml.property  name="fields"
	 */
    public String[] getFields() {
		String[] attributes = new String[fields.length - 1];
		for (int i = 1; i < fields.length; i++)
			attributes[i - 1] = fields[i];
		return attributes;
	}

    public String getTableName() {
        try {
			return layer.getQualifiedName();
		} catch (SeException e) {
			e.printStackTrace();
		}
		return layer.getName();
    }

    private synchronized SeRow obtainRow(long rowIndex) throws SeException {
       if (rowIndex != posActual) {
            if (rowIndex == 0) {
                if (query != null) {
                    if (query.inProgress()) {
                        query.close();
                    }
                    query = new SeQuery(((ConnectionSDE)spatialConnection).getConnection(), fields, sqlConstruct);
                    query.prepareQuery();
                    query.execute();
                }
                lastRowServed = -1;
            }

            if (lastRowServed == (rowIndex - 1)) {
                row = query.fetch();
                lastRowServed++;
            } else {
                if (queryAux != null && queryAux.inProgress()) {
                	queryAux.close();
                }

                queryAux = new SeQuery(((ConnectionSDE)spatialConnection).getConnection(), fields, sqlConstruct);
                SeObjectId rowID = new SeObjectId(rowIndex + 1);
                row = queryAux.fetchRow(layer.getQualifiedName(), rowID, fields);
            }
            posActual = rowIndex;
            cachedRow = row;
        }

        return cachedRow;
    }

    public Value getFieldValue(long rowIndex, int fieldId)
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
        SeRow row;
        Value val = null;

        try {
            int idFieldArcSDE = fieldId+1; // SIEMPRE CONTANDO CON
            // QUE NOS HAN PASADO EL PRIMER CAMPO EL DE SHAPE
            row = obtainRow(rowIndex);
            if (row==null)
            	return null;
            int dataType = colDefs[idFieldArcSDE].getType();

            switch (dataType) {
            case SeColumnDefinition.TYPE_INT16:
                val = ValueFactory.createValue(row.getShort(idFieldArcSDE)
                                                  .intValue());

                break;

            case SeColumnDefinition.TYPE_INT32:
            case SeColumnDefinition.TYPE_INT64:
                val = ValueFactory.createValue(row.getInteger(idFieldArcSDE)
                                                  .intValue());

                break;

            case SeColumnDefinition.TYPE_FLOAT32:
                val = ValueFactory.createValue(row.getFloat(idFieldArcSDE)
                                                  .floatValue());

                break;

            case SeColumnDefinition.TYPE_FLOAT64:
                val = ValueFactory.createValue(row.getDouble(idFieldArcSDE)
                                                  .doubleValue());

                break;

            case SeColumnDefinition.TYPE_DATE:
                val = ValueFactory.createValue(row.getTime(idFieldArcSDE)
                                                  .getTime());

                break;

            case SeColumnDefinition.TYPE_STRING:
                String strAux = row.getString(idFieldArcSDE);

                if (strAux == null) {
                    strAux = "";
                }

                val = ValueFactory.createValue(strAux);

                break;
            case SeColumnDefinition.TYPE_NSTRING:
                String nstrAux = row.getNString(idFieldArcSDE);
                if (nstrAux == null) {
                    nstrAux = "";
                }

                val = ValueFactory.createValue(nstrAux);

                break;
            }

            return val;
        } catch (SeException e) {
        	NotificationManager.addError(e);
        }

        return ValueFactory.createNullValue();
    }

    public int getFieldCount()
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
        return fields.length - 1;
    }

    public String getFieldName(int fieldId)
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
        return fields[fieldId + 1];
    }

    public long getRowCount() {
        return numReg;
    }

    public int getFieldType(int i)
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
        return fieldTypes[i+1];
    }

    /**
     * @see com.hardcode.gdbms.engine.data.driver.ObjectDriver#getPrimaryKeys()
     */
    public int[] getPrimaryKeys()
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
        throw new UnsupportedOperationException();
    }

    /**
     * @see com.hardcode.gdbms.engine.data.driver.ObjectDriver#write(com.hardcode.gdbms.engine.data.edition.DataWare)
     */
    public void write(DataWare arg0)
        throws com.hardcode.gdbms.engine.data.driver.DriverException {
    }

    public IFeatureIterator getFeatureIterator(Rectangle2D r, String strEPSG,
        String[] alphaNumericFieldsNeeded) throws DriverException {
        /*
        *   Generate a rectangular shape that will be used as a filter
        */
        try {
            SeShape shape = new SeShape(layer.getCoordRef());
            SeExtent extent=layer.getExtent();
            shape.generateRectangle(extent);

            SeShape[] shapes = new SeShape[1];
            shapes[0] = shape;

            /*
             *   Retrieve all the shapes that are contained within the
             *   rectangles envelope.
             */
            SeShapeFilter[] filters = new SeShapeFilter[1];
            SeShapeFilter filter = null;

            filter = new SeShapeFilter(layer.getQualifiedName(),
                    layer.getSpatialColumn(), shape, SeFilter.METHOD_ENVP);
            filters[0] = filter;

            SeQuery spatialQuery = null;
            SeSqlConstruct sqlCons = new SeSqlConstruct(layer.getQualifiedName());

            spatialQuery = new SeQuery(alphanumericConnection, fields, sqlCons);

            /*
             *   Set spatial constraints
             */
            spatialQuery.setSpatialConstraints(SeQuery.SE_OPTIMIZE, false,
                filters);

            spatialQuery.prepareQuery();

            spatialQuery.execute();

            return new ArcSdeFeatureIterator(spatialQuery,getLyrDef().getFieldID());
        } catch (SeException e) {
        	NotificationManager.addError(e);

            return null;
        }
    }

    public boolean isWritable() {
        return false;
    }
    public void setData(IConnection connection, DBLayerDefinition lyrDef) throws DBException {
    	 String[] fieldsAux=null;
         FieldDescription[] fds=lyrDef.getFieldsDesc();
    	 if ((fds != null) && (fds.length != 0)) {
             if (!fds[0].getFieldName().equals(lyrDef.getFieldGeometry())) {
                 fieldsAux = new String[fds.length + 1];
                 fieldsAux[0] = lyrDef.getFieldGeometry();

                 for (int k = 1; k < fds.length+1; k++) {
                     fieldsAux[k] = fds[k - 1].getFieldName();
                 }
             }else{
            	 fieldsAux = new String[fds.length];
            	 for (int k = 0; k < fds.length; k++) {
                     fieldsAux[k] = fds[k].getFieldName();
                 }
             }
         }
        lyrDef.setFieldNames(fieldsAux);
        lyrDef.setIdFieldID(lyrDef.getIdField(lyrDef.getFieldID())-1);
        setLyrDef(lyrDef);
        // Conexin:
        String server = lyrDef.getHost();
        int instance = lyrDef.getPort();
        String database = lyrDef.getDataBase();
        String user = lyrDef.getUser();
//        String password = lyrDef.getPassword();
        String[] cols = lyrDef.getFieldNames();
        String tableName = lyrDef.getComposedTableName();
        String password;
        if (lyrDef.getPassword()!=null){
        	password=lyrDef.getPassword();
        }else{
        	String keyPool = getName().toLowerCase() + "_" + lyrDef.getHost().toLowerCase()
        		+ "_" + lyrDef.getPort() + "_" + lyrDef.getDataBase().toLowerCase()
        		+ "_" + lyrDef.getUser().toLowerCase();
        	password=(String)poolPassw.get(keyPool);
        }
        try {
            alphanumericConnection =new SeConnection(server, instance, database, user, password);
//            alphanumericConnection.startTransaction();
            spatialConnection = connection;
//            ((ConnectionSDE)spatialConnection).getConnection().startTransaction();
        } catch (SeException e) {
        	NotificationManager.addError(e);
            return;
		}
        // Fetching data
        String layerName = tableName;
        SeObjectId layerID = null;
        String strSpatialColumn = "";

        try {
            Vector theLayers = ((ConnectionSDE)spatialConnection).getConnection().getLayers();

            for (int i = 0; i < theLayers.size(); i++) {
                SeLayer layer = (SeLayer) theLayers.elementAt(i);

                if (layer.getQualifiedName().compareToIgnoreCase(layerName) == 0) {
                	layerID = layer.getID();
                    strSpatialColumn = layer.getSpatialColumn();
                    this.layer=layer;
                    break;
                }


            }

            if (layerID == null) {
                System.err.println("Capa no encontrada");

                return;
            }

            // layerName = layer.getName();
//            layer = new SeLayer(((ConnectionSDE)spatialConnection).getConnection(), layerName, strSpatialColumn);

            //TODO aqu se puede saber la proyeccin
            //layer.getCoordRef().getSrid();

//            if ((cols != null) && (cols.length != 0)) {
//                if (!cols[0].equals(lyrDef.getFieldGeometry())) {
//                    this.fields = new String[cols.length + 1];
//                    this.fields[0] = lyrDef.getFieldGeometry();
//
//                    for (int i = 1; i < fields.length; i++) {
//                        fields[i] = cols[i - 1];
//                    }
//                } else {
//                    fields = cols;
//                }
//            } else {
//                SeSqlConstruct sqlConstruct = new SeSqlConstruct(tableName);
//                sqlTotal = sqlConstruct.getWhere();
//
//                SeTable table1 = new SeTable(conn, tableName);
//
//                /*
//                 *   Get the table's column definition.
//                 */
//                SeColumnDefinition[] tableDef = table1.describe();
//                this.fields = new String[tableDef.length];
//
//                /*
//                 *   Store the names of all the table's columns in the
//                 *   String array cols. This array specifies the columns
//                 *   to be retrieved from the database.
//                 */
//                int idField = 1;
//
//                for (int i = 0; i < tableDef.length; i++) {
//                    if (tableDef[i].getType() == SeColumnDefinition.TYPE_SHAPE) {
//                        this.fields[0] = tableDef[i].getName();
//                    } else {
//                        this.fields[idField] = tableDef[i].getName();
//                        idField++;
//                    }
//                }
//
//                lyrDef.setFieldNames(fields);
//            }
            fields=cols;
            sqlConstruct = new SeSqlConstruct(layerName);

            //      Create a query stream between the client and server
            query = new SeQuery(alphanumericConnection, fields, sqlConstruct);
            query.prepareQuery();
            query.execute();
            row = query.fetch();
            if (row == null) {
                System.out.println("La tabla " + getName() +
                    " no tiene registros");

                return;
            }

            colDefs = row.getColumns();
            lyrDef.setShapeType(obtainShapeType(row.getColumnDef(0).getType()));
            fieldTypes = new int[colDefs.length];

            for (int colNum = 0; colNum < colDefs.length; colNum++) {
                SeColumnDefinition colDef = colDefs[colNum];
                int dataType = colDef.getType();

                switch (dataType) {
                case SeColumnDefinition.TYPE_SHAPE:

                    if (colNum != 0) {
                        System.err.println(
                            "Por favor, el campo de shapes deber ser el primero");

                        return;
                    }

                    idSpatialColumn = colNum;

                    break;

                case SeColumnDefinition.TYPE_INT16:
                case SeColumnDefinition.TYPE_INT32:
                case SeColumnDefinition.TYPE_INT64:
                    fieldTypes[colNum] = Types.INTEGER;

                    break;

                case SeColumnDefinition.TYPE_FLOAT32:
                    fieldTypes[colNum] = Types.FLOAT;

                    break;

                case SeColumnDefinition.TYPE_FLOAT64:
                    fieldTypes[colNum] = Types.DOUBLE;

                    break;

                case SeColumnDefinition.TYPE_DATE:
                    fieldTypes[colNum] = Types.DATE;

                    break;

                default:
                    fieldTypes[colNum] = Types.VARCHAR;

                    break;
                }
            }
            fields=cols;
            SeQuery extentQuery = new SeQuery(((ConnectionSDE)spatialConnection).getConnection(), fields, sqlConstruct);
            SeQueryInfo queryInfo = new SeQueryInfo();
            queryInfo.setConstruct(sqlConstruct);

            SeExtent seExtent = extentQuery.calculateLayerExtent(queryInfo);
            fullExtent = new Rectangle2D.Double(seExtent.getMinX(),
                    seExtent.getMinY(),
                    seExtent.getMaxX() - seExtent.getMinX(),
                    seExtent.getMaxY() - seExtent.getMinY());
            extentQuery.close();

            doRelateID_FID();
        } catch (SeException e) {
        	throw new DBException(e);
        }
    }

    private int obtainShapeType(int type) {
        int shapeType = -1;

        switch (type) {
        case SeLayer.TYPE_POINT:
            shapeType = FShape.POINT;

            break;

        case SeLayer.TYPE_LINE:
        case SeLayer.TYPE_SIMPLE_LINE:
        case SeLayer.TYPE_MULTI_LINE:
        case SeLayer.TYPE_MULTI_SIMPLE_LINE:
            shapeType = FShape.LINE;

            break;

        case SeLayer.TYPE_POLYGON:
        case SeLayer.TYPE_MULTI_POLYGON:
            shapeType = FShape.POLYGON;

            break;

        default:
            shapeType = FShape.MULTI;
        }

        return shapeType;
    }

    public String getCompleteWhere() {
        return null;
    }

    public String getConnectionStringBeginning() {
        return "sde";
    }

    public int getDefaultPort() {
        return 5151;
    }

    public String getSourceProjection(IConnection conn,DBLayerDefinition dbld) {
        return originalEPSG;
    }

    public String getDestProjection() {
        return strEPSG;
    }

    public void setDestProjection(String toEPSG) {
        this.strEPSG = toEPSG;
    }

    public boolean canReproject(String toEPSGdestinyProjection) {
        //		 TODO POR AHORA, REPROYECTA SIEMPRE gvSIG.
        return false;
    }

    public IWriter getWriter() {
    	 if (writer == null) {
             writer = new ArcSdeWriter();
             writer.setDriver(this);
             writer.setSeLayer(layer);
             try {
                 writer.initialize(getLyrDef());
             }
             catch (EditionException e) {

             }
         }

         return writer;
    }

    /**
	 * @return
	 * @uml.property  name="sqlTotal"
	 */
    public String getSqlTotal() {
		return sqlTotal;
	}

    public XMLEntity getXMLEntity() {
        DBLayerDefinition lyrDef = getLyrDef();
        XMLEntity xml = new XMLEntity();
        xml.putProperty("className", this.getClass().getName());
        xml.putProperty("driverclass", this.getClass().getName());
        xml.putProperty("catalog", lyrDef.getCatalogName());

        ConnectionWithParams cwp =
        	SingleVectorialDBConnectionManager.instance().findConnection(getConnection());


        // TODO: NO DEBEMOS GUARDAR EL NOMBRE DE USUARIO Y CONTRASEA
        // AQUI. Hay que utilizar un pool de conexiones
        // y pedir al usuario que conecte a la base de datos
        // en la primera capa. En el resto, usar la conexin
        // creada con anterioridad.
        String userName = lyrDef.getUser();
        if (userName != null) {
            int aux = userName.indexOf("@");

            if (aux != -1) {
                userName = userName.substring(0, aux);
            }
        }

        xml.putProperty("username", userName);

        xml.putProperty("tablename", getTableName());
        xml.putProperty("fields", fields);
        xml.putProperty("ID", lyrDef.getFieldID());
        xml.putProperty("idfield",lyrDef.getIdFieldID());
        xml.putProperty("THE_GEOM", lyrDef.getFieldGeometry());
        xml.putProperty("whereclause", getWhereClause());
        xml.putProperty("SRID", lyrDef.getSRID_EPSG());

        xml.putProperty("host", lyrDef.getHost());
        xml.putProperty("port", lyrDef.getPort());
        xml.putProperty("dbName", lyrDef.getDataBase());
        xml.putProperty("connName", cwp.getName());

        xml.putProperty("typeShape",lyrDef.getShapeType());

        if (getWorkingArea() != null) {
            xml.putProperty("minXworkArea", getWorkingArea().getMinX());
            xml.putProperty("minYworkArea", getWorkingArea().getMinY());
            xml.putProperty("HworkArea", getWorkingArea().getHeight());
            xml.putProperty("WworkArea", getWorkingArea().getWidth());
        }

        return xml;
    }

    public void setXMLEntity(XMLEntity xml) throws XMLException {
        className = xml.getStringProperty("className");
        catalogName = xml.getStringProperty("catalog");
        lyrDef = new DBLayerDefinition();

        if (xml.contains("username")) {
            lyrDef.setUser(xml.getStringProperty("username"));
        }

        driverClass = xml.getStringProperty("driverclass");
        lyrDef.setTableName(xml.getStringProperty("tablename"));
        fields = xml.getStringArrayProperty("fields");
        lyrDef.setFieldNames(fields);
        lyrDef.setFieldID(xml.getStringProperty("ID"));
        if (xml.contains("idfield"))
        	lyrDef.setIdFieldID(xml.getIntProperty("idfield"));
        lyrDef.setFieldGeometry(xml.getStringProperty("THE_GEOM"));
        lyrDef.setWhereClause(xml.getStringProperty("whereclause"));
        lyrDef.setSRID_EPSG(xml.getStringProperty("SRID"));

        if (xml.contains("host")) {
            lyrDef.setHost(xml.getStringProperty("host"));
            lyrDef.setPort(Integer.parseInt(xml.getStringProperty("port")));
            lyrDef.setDataBase(xml.getStringProperty("dbName"));
            lyrDef.setConnectionName(xml.getStringProperty("connName"));
        }
        lyrDef.setShapeType(xml.getIntProperty("typeShape"));
        if (xml.contains("minXworkArea")) {
            double x = xml.getDoubleProperty("minXworkArea");
            double y = xml.getDoubleProperty("minYworkArea");
            double H = xml.getDoubleProperty("HworkArea");
            double W = xml.getDoubleProperty("WworkArea");
            workingArea = new Rectangle2D.Double(x, y, W, H);
        }

        lyrDef.setCatalogName(catalogName);

        if (workingArea != null) {
            lyrDef.setWorkingArea(workingArea);
        }

        setLyrDef(lyrDef);
    }

    public int getFieldWidth(int fieldId) {
        return colDefs[fieldId+1].getSize();
    }

    public void setLyrDef(DBLayerDefinition lyrDef) {
        this.lyrDef =lyrDef;
    }

    /**
	 * @see  com.iver.cit.gvsig.fmap.drivers.IVectorialJDBCDriver#setWorkingArea(java.awt.geom.Rectangle2D)
	 * @uml.property  name="workingArea"
	 */
    public void setWorkingArea(Rectangle2D rect) {
		this.workingArea = rect;
	}

    /**
	 * @see  com.iver.cit.gvsig.fmap.drivers.IVectorialJDBCDriver#getWorkingArea()
	 * @uml.property  name="workingArea"
	 */
    public Rectangle2D getWorkingArea() {
		return workingArea;
	}

    /**
     * @see com.iver.cit.gvsig.fmap.layers.ReadableVectorial#getShapeType()
     */
    public int getShapeType() {
        return lyrDef.getShapeType();
    }

    /* (non-Javadoc)
     * @see com.iver.cit.gvsig.fmap.drivers.VectorialDatabaseDriver#getWhereClause()
     */
    public String getWhereClause() {
        if (lyrDef.getWhereClause() == null) {
            return "";
        }

        return lyrDef.getWhereClause().toUpperCase();
    }

    /**
	 * @return
	 */
    public DBLayerDefinition getLyrDef() {
    	// Esto hace saltar un error.
//    	if (this.spatialConnection != null){
//    		lyrDef.setSeConnection();
//    	}
        return lyrDef;
    }

    /**
     * Empty method called when the layer is going to be removed from the view.
     * Subclasses can overwrite it if needed.
     */
    public void remove() {
    }

    /* (non-Javadoc)
    * @see com.iver.cit.gvsig.fmap.drivers.VectorialDriver#reLoad()
    */
    public void reload() throws IOException {
    	try {
			load();
		} catch (DriverException e) {
			NotificationManager.addError(e);
		}
       	//setData(null, lyrDef);
    }

    public void setDataSourceFactory(DataSourceFactory dsf) {
        // TODO Auto-generated method stub
    }

    public void load() throws DriverException {
    	try {
            if (driverClass != null)
            	Class.forName(driverClass);

            String _drvName = getName();

            String keyPool = _drvName.toLowerCase() + "_" + lyrDef.getHost().toLowerCase()
            + "_" + lyrDef.getPort() + "_" + lyrDef.getDataBase().toLowerCase()
            + "_" + lyrDef.getUser().toLowerCase();

            IConnection newConn = null;
            String clave = null;
            ConnectionWithParams cwp = null;

            if (poolPassw.containsKey(keyPool)) {

                clave = (String) poolPassw.get(keyPool);
                cwp = SingleVectorialDBConnectionManager.instance().getConnection(
 					_drvName, lyrDef.getUser(), clave, lyrDef.getConnectionName(),
 					lyrDef.getHost(), String.valueOf(lyrDef.getPort()), lyrDef.getDataBase(), true);

            } else {

                cwp = SingleVectorialDBConnectionManager.instance().getConnection(
     					_drvName, lyrDef.getUser(), null, lyrDef.getConnectionName(),
     					lyrDef.getHost(), String.valueOf(lyrDef.getPort()), lyrDef.getDataBase(), false);

                if (cwp.isConnected()) {

                	poolPassw.put(keyPool, cwp.getPw());

                } else {

                    JPasswordDlg dlg = new JPasswordDlg();
                    dlg.setLocationRelativeTo((Component)PluginServices.getMainFrame());
                    String strMessage = Messages.getString("conectar_jdbc");
                    String strPassword = Messages.getString("password");
                    dlg.setMessage(strMessage
                    		+ " ["
                    		+ _drvName + ", "
                    		+ lyrDef.getHost() + ", "
                    		+ String.valueOf(lyrDef.getPort()) + ", "
                    		+ lyrDef.getDataBase() + ", "
                    		+ lyrDef.getUser() + "]. "
                    		+ strPassword
                    		+ "?");
                    dlg.setVisible(true);
                    clave = dlg.getPassword();
                    if (clave == null)
                        return;
                    poolPassw.put(keyPool, clave);

                    cwp.connect(clave);
                }
            }

            newConn = cwp.getConnection();
//            newConn.setAutoCommit(false);

            DBLayerDefinition lyrDef = new DBLayerDefinition();
            if (getLyrDef() == null) {
	            lyrDef.setCatalogName(catalogName);
	            lyrDef.setTableName(getTableName());
	            lyrDef.setFieldNames(fields);
	            lyrDef.setFieldID(FIDfield);
	            lyrDef.setFieldGeometry(geometryField);
	            lyrDef.setWhereClause(whereClause);
	            // lyrDef.setClassToInstantiate(driverClass);
	            if (workingArea != null)
	                lyrDef.setWorkingArea(workingArea);

	            lyrDef.setSRID_EPSG(strSRID);
            } else {
            	lyrDef = getLyrDef();
            }

            setData(newConn, lyrDef);
        } catch (ClassNotFoundException e) {
            DriverJdbcNotFoundExceptionType type =
            	new DriverJdbcNotFoundExceptionType();
            type.setDriverJdbcClassName(driverClass);
            type.setLayerName(this.getTableName());
            throw new DriverException("Driver JDBC no encontrado", e,  type);
        } catch (DBException e) {

        	throw new DriverException(e.getMessage());

		}
    }

    public IConnection getConnection() {
        return spatialConnection;
    }
	public String getConnectionString(
    		String host,
    		String port,
    		String dbname,
    		String user,
    		String pw) {

		String resp = getConnectionStringBeginning()+ ":" + host.toLowerCase();

		if (dbname.trim().length() > 0) {
			resp += ":" + port;
		} else {
			resp += ":" + getDefaultPort();
		}

		resp += ":" + dbname.toLowerCase();
		return resp;
    }
	public String[] getAllFields(IConnection conex, String tableName) throws DBException {
		try {
			SeTable table = null;
			Vector tables=((ConnectionSDE)conex).getConnection().getTables(SeDefs.SE_SELECT_PRIVILEGE);
			for (int i=0;i<tables.size();i++) {
				if (tableName.equals(((SeTable)tables.get(i)).getQualifiedName())) {
					table=(SeTable)tables.get(i);
					break;
				}
			}
			SeRegisteredColumn[] columns=table.getColumnList();
			String[] fieldNames=new String[columns.length];
			for (int i=0; i < columns.length; i++){
				fieldNames[i]=columns[i].getName();
			}
			return fieldNames;
		} catch (SeException e) {
			throw new DBException(e);
		}
	}
	public String[] getAllFieldTypeNames(IConnection conex, String tableName) throws DBException {
		try {
			SeTable table = null;
			Vector tables=((ConnectionSDE)conex).getConnection().getTables(SeDefs.SE_SELECT_PRIVILEGE);
			for (int i=0;i<tables.size();i++) {
				if (tableName.equals(((SeTable)tables.get(i)).getQualifiedName())) {
					table=(SeTable)tables.get(i);
					break;
				}
			}
			SeRegisteredColumn[] columns=table.getColumnList();
			String[] fieldNames=new String[columns.length];
			for (int i=0; i < columns.length; i++){
				fieldNames[i]=String.valueOf(columns[i].getType());
			}
			return fieldNames;
		} catch (SeException e) {
			throw new DBException(e);
		}
	}
	public String[] getIdFieldsCandidates(IConnection conn2, String tableName) throws DBException {
		return getAllFields(conn2,tableName);
	}
	public String[] getGeometryFieldsCandidates(IConnection conn2, String tableName) {
		return new String[]{"SHAPE"};//getAllFields(conn2,tableName);
	}
	public String[] getTableNames(IConnection conex, String dbName) throws DBException {
		 try {
			Vector theLayers = ((ConnectionSDE)conex).getConnection().getLayers();
			TreeMap ret = new TreeMap();
			for (int i = 0; i < theLayers.size(); i++) {
				SeLayer layer = (SeLayer) theLayers.elementAt(i);
				ret.put(layer.getQualifiedName(), layer.getQualifiedName());
			}
			return (String[]) ret.keySet().toArray(new String[0]);
			} catch (SeException e) {
				throw new DBException(e);
			}
		}
	 public String[] getTableFields(IConnection conex,String tableName) throws DBException{
			try {
				SeTable table = new SeTable(((ConnectionSDE)conex).getConnection(),tableName);
				SeRegisteredColumn[] columns=table.getColumnList();
				String[] fieldNames=new String[columns.length];
				for (int i=0; i < columns.length; i++){
					fieldNames[i]=columns[i].getName();
				}
				return fieldNames;
			} catch (SeException e) {
				throw new DBException(e);
			}
		}

}
