package com.iver.cit.gvsig.fmap.layers;

import org.cresques.cts.IProjection;

import com.hardcode.driverManager.IDelayedDriver;
import com.hardcode.gdbms.engine.instruction.FieldNotFoundException;
import com.iver.cit.gvsig.fmap.DBDriverExceptionType;
import com.iver.cit.gvsig.fmap.DriverException;
import com.iver.cit.gvsig.fmap.GenericDriverExceptionType;
import com.iver.cit.gvsig.fmap.LegendDriverExceptionType;
import com.iver.cit.gvsig.fmap.drivers.WithDefaultLegend;
import com.iver.cit.gvsig.fmap.drivers.sde.IVectorialSDEDriver;
import com.iver.cit.gvsig.fmap.layers.DefaultDelayedDriverListener;
import com.iver.cit.gvsig.fmap.rendering.LegendFactory;
import com.iver.cit.gvsig.fmap.rendering.VectorialLegend;
import com.iver.utiles.ExceptionDescription;

public class SDELayerFactory {
public static SDELayer getLayerSDE(IVectorialSDEDriver driver,String layerName,IProjection proj){
	SDELayer capa = new SDELayer();

	capa.setName(layerName);
	VectorialSDEAdapter sdeAdapter = new VectorialSDEAdapter();
	sdeAdapter.setDriver(driver);

	capa.setSource(sdeAdapter);
	capa.setProjection(proj);
	try {
		if (driver instanceof WithDefaultLegend) {
			WithDefaultLegend aux = (WithDefaultLegend) driver;
			sdeAdapter.start();
			capa.setLegend((VectorialLegend) aux.getDefaultLegend());
			sdeAdapter.stop();
		} else {
			capa.setLegend(LegendFactory.createSingleSymbolLegend(capa
					.getShapeType()));
		}
		if (driver instanceof IDelayedDriver)
		{
			// Por defecto, los drivers estn listos para entregar
			// features al terminar su initialize. Pero con los drivers
			// que implementan IDelayedDriver, el driver es responsable
			// de avisar cundo est listo
			capa.getFLayerStatus().setDriverLoaded(false);
			((IDelayedDriver) driver).addDriverEventListener(new DefaultDelayedDriverListener(capa));
		}
		// PARCHE: Llamamos a getXmlEntity del driver para que se rellenen
		// los valores de host, port, etc, necesarios para un posible reload
		//FIXME: Se vuelve a dejar como estaba... el problema esta en el DefaultDBDriver.getXMLEntiy()
		//       y comentarizemos esto o no, si no arreglamos ese metodo da problemas de persistencia
		driver.setXMLEntity(driver.getXMLEntity());

	} catch (FieldNotFoundException e) {
		LegendDriverExceptionType excepType =
			new LegendDriverExceptionType("Error al construir la leyenda, campo no encontrado");
		//TODO Para hacer esto extensible tiene que usarse puntos
		//de extension, y comparar las clases de leyendas registradas
		VectorialLegend legend = (VectorialLegend)
			((WithDefaultLegend)driver).getDefaultLegend();

		excepType.setLegendLabelField(legend.getLabelField());
		excepType.setLegendHeightField(legend.getLabelHeightField());
		excepType.setLegendRotationField(legend.getLabelRotationField());
		DriverException exception = new DriverException(e, excepType);
		capa.setAvailable(false);
		capa.addError(exception);
		return capa;
		// throw new UnsupportedOperationException(e.getMessage());
	} catch (DriverException e) {
		DBDriverExceptionType excepType =
			new DBDriverExceptionType();
		excepType.setDriverName(driver.getName());
		excepType.setTableName(driver.getTableName());
		DriverException exception = new DriverException(e, excepType);
		capa.addError(exception);
		capa.setAvailable(false);
		return capa;
		// throw new UnsupportedOperationException(e.getMessage());
	} catch (Exception e) {
		ExceptionDescription excepType = new GenericDriverExceptionType();
		DriverException exception = new DriverException(e, excepType);
		capa.addError(exception);
		capa.setAvailable(false);
		return capa;
	}

	return capa;
}

}
